unit miniini;
(*******************************************************
   Minimum ini file
   all lines are key=value with no frills
   allow ';' to indicate comments


   Use
   --------------------------------------------------------------------
   (1) Declare  M:TminiIni with enough public visibility for your needs
   (2) At start up     M := TminiIni.Create('');   // or similar will read data
   (3) At close down   M.free;                     // will save data automatically

   Filename can be
   * executable with extension changed to .conf
   * real file name
   * alias path + application executable + .conf
   See create().  Use .name property to check actual filename

   In code use the get.. and set.. methods

   The position of forms can be loaded/saved in one hit
   use something like LoadFormsettings(Tform(myForm));
   Set the position property of form to poDesigned.
   (See code for option to save/load font at same time)

   Fonts (name,size,bold,italic) can be loaded and saved


   Most recently used
   ------------------
   A frequent requirement is to store a MRU list.
   Use SetMru to stuff a filename onto the top of the list.
   GetMruList is used to collect all MRU items into a TStrings.
   This means you can do something like
     .getmrulist(listbox1.items)

   The number of items in the list is controlled by an entry 'MRUmax'.
   either
     .SetMru(LastUsedFileName,0);    // uses existing MRUmax or defaults to 4 if not set at all
     .SetMru(LastUsedFileName,10);   // does all the housekeeping and sets list to 10 items
     .setint('MRUmax',10);           // explicit setting but NO HOUSEKEEPING


   Special key values
   ------------------
   ~U~ and ~C~ will get replaced by User name and computer name respectively
   This can be used to save separate settings for different users etc.
   eg.  .GetDateTime('LastLogin~U~');  // when did user last logon

   if .get methods cannot find <key><u-name or c-name> they will look
   instead for just <key>

   StripOut method
   ---------------
   If a user leaves or a computer is replaced you will get left over entries
   where you have been using ~U~ and ~C~.  The stripout method removes all
   items with the specified search string.  eg .StripOut('FredSmith')


*******************************************************)


interface

uses     useful,Classes,sysutils,forms,MessageDialogBx,windows,graphics;

type
   TMiniIni = class(tstringlist)
     private
       FFilename : string;
       FLoaded   : boolean;
       FChanged  : boolean;
       FExtended : boolean;
       FFormExtension : string;
       FMruMax   : integer;
       procedure   Load;
       procedure   Save;
       Function    ExtendKey(const ky:string):string;
     public
       constructor create(const aFilename:string);  overload; virtual;
       constructor create(const aFilename,FormExtensionOptions:string);  overload;virtual;
       destructor  destroy;  override;
       function    getstr(const ky:string):string;                    virtual;
       function    getint(const ky:string):integer;
       function    getintdef(const ky:string;const deflt:integer):integer;
       function    getDateTime(const ky:string):TDatetime;
       function    getBool(const ky:string):boolean;
       function    getRect(const ky:string):Trect;           // returns Left=-1 if key not found
       function    GetFont(const ky:string):Tfont;
       procedure   setstr(const ky,v:string);                         virtual;
       procedure   setInt(const ky:string;  const v:integer);         virtual;
       procedure   setDateTime(const ky:string; const v:TDatetime);   virtual;
       procedure   setBool(const ky:string;  const v:boolean);        virtual;
       procedure   setRect(const ky:string;  const v:Trect);          virtual;
       procedure   SetFont(const ky:string;var fnt:TFont);
       procedure   LoadFormsettings(var f:Tform);
       procedure   SaveFormsettings(var f:Tform);
       procedure   flush;
       procedure   add(const s:string);
       procedure   StripOut(const srchkey:string);
       procedure   GetMruList(const sl:tstrings);
       procedure   SetMru(const fname:string; const mrumax:integer);
     published
       property    Changed : boolean read FChanged write Fchanged;
       property    name : string  read FFilename;
     end;


implementation


constructor TMiniIni.create(const aFilename,FormExtensionOptions:string);
{ The additional argument is ~U~, ~C~ or both or none
 which indicates that when form settings are loaded/saved the user or computer name
 is added to the key to personalise them.
}
begin
  create(aFilename);
  FFormExtension := FormExtensionOptions;
end;

constructor TMiniIni.create(const aFilename:string);
{ set up the mini-ini to look at the specified file
  if aFilename is blank then use the application executable with extension changed to '.conf'
  if aFilename is not blank but doesn't contain a '.' then interpret
  it as an alias and use the _path_ from that plus the application _name_ plus the extension '.conf' }
var
  f : string;
begin
  inherited create;
  // work out filename
  f := afilename;
  if f = '' then begin
    f := changefileext(application.exename,'.conf');
  end else begin
    if pos('.',f)=0 then begin
      f := pathforalias(f)+'\'+extractfilename(changefileext(application.exename,'.conf'));
    end;
  end;
  FFilename := f;
  if not fileexists(f) then begin
    okbx('Warning|The configuration file "'+f+'" cannot be found');
    Floaded := true;
    Fchanged := false;
  end else begin
    Load;
  end;
  FFormExtension := '';

end;

destructor TMiniIni.destroy;
{ make sure it is saved if any changes have been made }
begin
  if FChanged then save;
  inherited destroy;
end;

procedure  TMiniIni.flush;
{ make sure it is saved if any changes have been made }
begin
  if FChanged then save;
end;


procedure TMiniIni.load;
{ load from file given in FFilename }
begin
  if not Floaded then begin
    loadfromfile(FFilename);
    FLoaded := true;
    FChanged := false;
  end;
end;

procedure TMiniIni.save;
{ save to file given in FFilename }
begin
  if Floaded then begin
    if Fchanged then begin
      savetofile(FFilename);
      FChanged := false;
    end;
  end;
end;

procedure   Tminiini.add(const s:string);
{ raw add of a complete string }
begin
  inherited add(s);
  FChanged := true;
end;


function    TMiniIni.getstr(const ky:string):string;
{ this is the core 'read a value' method
  If the key is extended with ~U~ or ~C~ then extend and look for
  the value with the extension.  If no value can be found for the
  extension then default by looking for value without extension
}

var
  eky,v : string;
begin
  // possibly have a go at extended keys first
  eky := ExtendKey(ky);
  if FExtended then begin
    if indexofName(eky)>-1 then begin
      v := trim(values[eky]);
      result := trim(chop(v,';'));
      exit;
    end else begin
      eky := substr(substr(ky,'~C~',''),'~U~','');
    end;
  end;
  // plain key
  v := trim(values[eky]);
  result := trim(chop(v,';'));
end;


function    TMiniIni.getint(const ky:string):integer;
begin
  result := strtointdef(getstr(ky),0);
end;

function    TMiniIni.getintdef(const ky:string;const deflt:integer):integer;
begin
  result := getint(ky);
  if result = 0 then result := deflt;
end;

function    TMiniIni.getbool(const ky:string):boolean;
begin
  result := getstr(ky)='TRUE';
end;


function    TMiniIni.getDateTime(const ky:string):TDatetime;
{ if not found the return NotAValidDate }
var
  s : string;
begin
  s := getstr(ky);
  if s = '' then result := NotAValidDate
            else result := strtodatetime(s);
end;

function    TMiniIni.GetRect(const ky:string):Trect;
{ return a rectangle - eg for saving form positions.
  If the key isn't found then left is -1 and other values are
  set to a workable, but who knows? default}
var
  s : string;
  l,t,r,b : integer;
begin
  l := -1;           // defaults and not found
  t := 0;
  r := 600;
  b := 300;
  s := GetStr(ky);  // L|T|R|B
  if s <> '' then begin
    l := strtointdef(chop(s,'|'),l);
    t := strtointdef(chop(s,'|'),t);
    r := strtointdef(chop(s,'|'),r);
    b := strtointdef(s,l);
  end;
  result := rect(l,t,r,b);
end;



procedure   TMiniIni.setstr(const ky,v:string);
var
  eky,s,vold,vnew : string;
begin
  // always extend key first
  eky := ExtendKey(ky);
  s := values[eky];                           // full text including any ;
  vold := trim(chop(s,';'));                  // active element
  if vold<>v then begin
    vnew := v;
    if trim(s)<>'' then vnew := vnew + '  ;'+s;       // will preserve trailing comments like '; ....'
    values[eky] := vnew;
    FChanged := true;
  end;
end;


procedure   TMiniIni.setInt(const ky:string; const v:integer);
begin
  SetStr(ky,inttostr(v));
end;

procedure   TMiniIni.setBool(const ky:string; const v:boolean);
begin
  if v then SetStr(ky,'TRUE') else SetStr(ky,'FALSE');
end;


procedure   TMiniIni.setDateTime(const ky:string; const v:TDatetime);
begin
  SetStr(ky,formatdatetime('dd/mm/yy hh:nn:ss',v));
end;


procedure   TMiniIni.setRect(const ky:string;  const v:Trect);
var
  s : string;
  l : integer;
begin
  l := v.left;               // this hacking is done 'cos "l=-1" signals key not found.
  if l = -1 then l := 0;
  s := inttostr(l)+'|'+inttostr(v.top)+'|'+inttostr(v.right)+'|'+inttostr(v.bottom);
  SetStr(ky,s);
end;



procedure   TMiniIni.LoadFormsettings(var f:Tform);
{ Converse of SaveFormSettings  q.v.
  Do nothing if keys are not found
}
var
  r : Trect;
  ky : string;
begin
  ky := f.name+FFormExtension;
  // position
  r := getRect(ky+'RECT');
  if r.left = -1 then exit;       // key has not been found so quit
  f.left := r.left;
  f.top := r.top;
  f.width := r.right-r.left;
  f.height := r.bottom - r.top;
  //  if GetStr(ky+'FONT')<>'' then begin
  //    f.font.assign(GetFont(ky+'FONT'));
  //  end;
end;


procedure   TMiniIni.SaveFormsettings(var f:Tform);
{ save the position for the form
  use the form's name as the main key}
var
  ky : string;
begin
  ky := f.name+FFormExtension;
  SetRect(ky+'RECT',Rect(f.left,f.top,f.left+f.width,f.top+f.height));
  //  SetFont(Ky+'FONT',f.font);
end;

function    TMiniIni.GetFont(const ky:string):TFont;
{ get the font properties (name,size,colour,italic,bold)
  return nil if not found
  call something like
  |  fnt := LoadFont('SomeKey');
  |  if assigned(fnt) then begin
  |    MyRealFont.assign(fnt);
  |    fnt.free;
  |  end;
}
var
  s : string;
begin
  s := GetStr(ky);
  if (s='') then begin
    result :=nil;
  end else begin
    result := Tfont.create;
    StringToFont(s,Result);
  end;
end;

procedure   TMiniIni.SetFont(const ky:string;var fnt:TFont);
{ save the font properties (name,size,colour,italic,bold) }
begin
  SetStr(ky,fonttostring(fnt));
end;


Function    TMiniIni.ExtendKey(const ky:string):string;
{ possibly substitute ~U~ by username and ~C~ by computer name

  set the expanded flag to show that there was a substitution
  which we need if we can't find what we want with the extension

}
begin
  result := ky;
  Fextended := pos('~',ky)>0;
  if not Fextended then exit;   // nothing to do
  if pos('~U~',result)<>0 then result := substr(result,'~U~',':'+NameOfUser);
  if pos('~C~',result)<>0 then result := substr(result,'~C~',':'+NameOfComputer);
end;

procedure   TMiniIni.StripOut(const srchkey:string);
{ remove any entries where srchkey is a partial or complete match with key
  Users and computers will be prefixed by : for more security
  NB: Key extension is NOT used so StripOut('Setting~U~') WILL NOT WORK
  To get the current user use NameOfUser function from useful.pas
   }
var
  i : integer;
  k : string;
begin
  k := trim(srchkey);
  if length(k)<2 then exit;                // 1 char is prob an error. 0 is silly
  for i := count-1 downto 0 do begin
    if pos(k,names[i])>0 then delete(i);
  end;
end;


procedure TMiniIni.GetMruList(const sl:tstrings);
{ read most recently used items into specified list }
var
  i : integer;
  s : string;
begin
  // what is the max no of entries
  FMruMax := getint('MRUmax');
  // special case - first time default to 4
  if FMruMax < 1 then begin
    FMruMax := 4;
    setint('MRUmax',FMruMax);
  end;
  // load into the list
  sl.clear;
  for i := 0 to FMruMax-1 do begin
    s := getstr('MRU'+inttostr(i));
    if s <> '' then sl.add(s);
  end;
end;


procedure TMiniIni.SetMru(const fname:string; const mrumax:integer);
{ add the file to the top of the MRU list
  MRUMAX is the max number of MRU entries to hold
  If mrumax < 1 then use existing
  ** NB your app may need to call GetMruList afterwards to update
  ** its display.

  If you want to clear the MRU list completely then do the following
    .StripOut('MRU');  // also clear's no of list items to be stored

}
var
  s : string;
  mm,i : integer;
  mrul : tstringlist;
begin

  mrul := tstringlist.create;
  TRY
    // load the list
    GetMruList(tstrings(mrul));

    // how many are we going to keep
    mm := mrumax;
    if mm < 1 then mm := FMruMax;

    // insert at top (possibly delete first)
    i := mrul.indexof(fname);
    if i<>-1 then mrul.delete(i);
    mrul.insert(0,fname);

    // write out what we have
    for i := 0 to mrumax-1 do begin
      if i>mrul.count-1 then s := '' else s := mrul.strings[i];
      setstr('MRU'+inttostr(i),s);
    end;

    // delete any left overs if we are reducing mrumax
    if mm < FMruMax then begin
      for i := mm to FMruMax-1 do begin
        setstr('MRU'+inttostr(i),'');
      end;
    end;

    // update mrumax
    if mm <> FMruMax then setint('MRUmax',mm);

  FINALLY
    mrul.free;
  END;

end;



end.
